/*

********************************************************************************
**      ATI Mach 64 chipset driver
********************************************************************************
**
    Copyright (C)       1998    Isaac Connor
                                Mathew Wilkins

********************************************************************************
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
*/

/* mach64_detect.c
 * 	Contains code for the detection of Mach64 based video cards.
 */

#ifndef _MACH64_DETECT_C__
#define _MACH64_DETECT_C__

#include "kgi/module.h"

#include "ATI/mach64.h"
#include "ATI/mach64_io.h"
#include "ATI/mach64_detect.h"


int mach64_detect( struct kgi_display* dpy ) {

  return pci_detect(dpy);

} /* mach64_detect( void ) */


/*--------
 * 
 *
 */

int pci_detect( struct kgi_display* dpy ) {
    int		status, id_num, temp, count;
    char 	bus, devfn;
    struct cardInfo *card;
    /*
     * check for pci bios
     */

    if( !pcibios_present() ) {
	printk( KERN_ERR "PCI Bus not detected.\n" );
	return -E( CHIPSET, UNKNOWN );
    }

    /*
     * check for all the mach64 codes
     */

    /* FIXME all pci-related linux system calls should be replaced with
       GGI ones (such as  pcicfg_in32() in system.h) */
    
    
    for ( id_num = 0; id_num < MACH64_UNKNOWN; id_num += 1 ) {

        /* The only time we will do multiple executions of this inner
           loop is when we have more than 1 of a particular card. */
        for ( count = 0; ; count += 1 ) {
            status = pcibios_find_device( PCI_VENDOR_ID_ATI,
				Mach64_PCI_Ids[ id_num ],
				count, &bus, &devfn );
                
	    if( PCIBIOS_SUCCESSFUL == status ) {
                /* Found a card, now fill in the structure for it */ 
#ifdef DEBUG
        printk( KERN_ERR "Mach64 detected: %s\n", Mach64_Descriptions[ id_num ]);
#endif
                    card = (struct cardInfo *)kgi_malloc_small(sizeof(struct cardInfo));

#ifdef DEBUG
   if ( ! card ) 
        printk( KERN_ERR "Error allocating memory.\n");
#endif

	        (struct cardInfo*)KGIM_PRIVATE(dpy, chipset) = card;

                card->Card_Type	= id_num;
                card->Bus	= bus;
                card->Devfn	= devfn;
                /*
                 * find out or guess mmio_base, sio_base, ram_size, dac_type
                 */

                pcibios_read_config_dword( bus, devfn, PCI_REVISION_ID, &temp );
 
                card->Revision	= temp & 255;

                pcibios_read_config_dword( bus, devfn, PCI_BASE_ADDRESS_0, &temp );

                card->LFB_Base	= temp & ~255;

                if ( id_num < MACH64_GX || id_num >= MACH64_UNKNOWN ) {
                    printk( KERN_ERR "Mach64: Unknown Card sub type : %d :\n",id_num );
                    return -E( CHIPSET, UNKNOWN );
                }


                /* How to tell if we have a newer card where MMIO is forced... */
                pcibios_read_config_dword( bus, devfn, PCI_BASE_ADDRESS_1, &temp );
                if ( temp & 0xfffffffc ) {
                    /* MMIO ON! Note: This seems to work...*/
                    card->Using_MMIO = 1;
                    card->Registers = (int *)&Mach64_MMIO_Registers;
                    card->SIO_Base = temp & ( temp & 0x1 ? 0xfffffffc : 0xfffffff0 );
                } else {
                    card->Using_MMIO = 0; 
                    card->Registers = (int *)&Mach64_Registers;
                    pcibios_read_config_dword( bus, devfn, PCI_STATUS_UDF, &temp );
                    card->SIO_Base     = Mach64_SIO_Bases[ ( temp >> 6 ) & 0x03 ];
                }
		
                get_registers( dpy);
                detect_ramsize( dpy );
                detect_aperture( dpy );
                detect_dac( dpy );


            } else {
                /* Move on to next id_num */
                break;
            } /* end if */
        } /* end for */
    } /* end for */

        printk( KERN_ERR "Mach64 detected: %s, rev %d, LFB Base 0x%8x.\n",
                  Mach64_Descriptions[ card->Card_Type ],
                  card->Revision & 255,
                  card->LFB_Base );
    return EOK;
}


/* Get Important Registers */
void get_registers( struct kgi_display *dpy ) {

	GET_PRIVATE;    

	printk("doing read/write test...\n");
        mach64_out32(Scratch_Reg1, 0x55555555, dpy);	
	printk("wrote 0x55555555, got back 0x%8x.\n",
	mach64_in32(Scratch_Reg1, dpy ));
			
        card->crtc_h_total       = mach64_in32( Crtc_H_Total, dpy );
        card->crtc_h_sync        = mach64_in32( Crtc_H_Sync, dpy );
        card->crtc_v_total       = mach64_in32( Crtc_V_Total, dpy );
        card->crtc_v_sync        = mach64_in32( Crtc_V_Sync, dpy );
        card->crtc_vline         = mach64_in32( Crtc_Vline, dpy );
        card->crtc_display_start = mach64_in32( Crtc_Display_Start, dpy );
        card->crtc_int_cntl      = mach64_in32( Crtc_Int_Cntl, dpy );
        card->crtc_gen_cntl      = mach64_in32( Crtc_Gen_Cntl, dpy );
        card->clock_cntl         = mach64_in32( Clock_Cntl, dpy  );
        card->bus_cntl           = mach64_in32( Bus_Cntl, dpy );
        card->mem_cntl           = mach64_in32( Mem_Cntl, dpy  );
        card->dac_cntl           = mach64_in32( Dac_Cntl, dpy  );
        card->gen_test_cntl      = mach64_in32( Gen_Test_Cntl, dpy  );
        card->config_cntl        = mach64_in32( Config_Cntl, dpy    );


#ifdef DEBUG
#warning Including Register Dump Code
        printk( KERN_ERR "Mach64: Register Dump.\n" );

        printk( KERN_ERR "  crtc_h_total  0x%8x.\n",    card->crtc_h_total );
        printk( KERN_ERR "  crtc_h_sync   0x%8x.\n",    card->crtc_h_sync );
        printk( KERN_ERR "  crtc_v_total  0x%8x.\n",    card->crtc_v_total );
        printk( KERN_ERR "  crtc_v_sync   0x%8x.\n",    card->crtc_v_sync );
        printk( KERN_ERR "  crtc_vline    0x%8x.\n",    card->crtc_vline );
        printk( KERN_ERR "  crtc_display_start 0x%8x.\n", card->crtc_display_start );
        printk( KERN_ERR "  crtc_int_cntl 0x%8x.\n",    card->crtc_int_cntl );
        printk( KERN_ERR "  crtc_gen_cntl 0x%8x.\n",    card->crtc_gen_cntl );
        printk( KERN_ERR "  clock_cntl    0x%8x.\n",    card->clock_cntl );
        printk( KERN_ERR "  bus_cntl      0x%8x.\n",    card->bus_cntl );
        printk( KERN_ERR "  mem_cntl      0x%8x.\n",    card->mem_cntl );
        printk( KERN_ERR "  dac_cntl      0x%8x.\n",    card->dac_cntl );
        printk( KERN_ERR "  config_cntl   0x%8x.\n",    card->config_cntl );
        printk( KERN_ERR "  gen_test_cntl 0x%8x.\n",    card->gen_test_cntl );

#endif

}


/*
 *	Determine the size of RAM
 */

void detect_ramsize( struct kgi_display *dpy ) {
    GET_PRIVATE;
    
    if ( 
        ( ( MACH64_GT == card->Card_Type ) && ( card->Revision & 0x07 ) ) || 
        ( MACH64_VT == card->Card_Type ) ||  
        ( MACH64_VU == card->Card_Type ) ||  
        ( MACH64_GU == card->Card_Type ) ||  
        ( MACH64_GB == card->Card_Type ) ||  
        ( MACH64_GD == card->Card_Type ) ||  
        ( MACH64_GI == card->Card_Type ) ||  
        ( MACH64_GP == card->Card_Type ) ||  
        ( MACH64_GQ == card->Card_Type )   
       ) {

        switch( card->mem_cntl & 0xf ) {
            case 0: card->Ram_Size = Mach64_Chip_Info.ramsize = 512 KB; break;
            case 1: card->Ram_Size = Mach64_Chip_Info.ramsize = 1 MB; break;
            case 3: card->Ram_Size = Mach64_Chip_Info.ramsize = 2 MB; break;
            case 7: card->Ram_Size = Mach64_Chip_Info.ramsize = 4 MB; break;
            case 9: card->Ram_Size = Mach64_Chip_Info.ramsize = 6 MB; break;
            case 0xb: card->Ram_Size = Mach64_Chip_Info.ramsize = 8 MB; break;
            case 0xf: card->Ram_Size = Mach64_Chip_Info.ramsize = 16 MB; break;
            default:
                printk( KERN_ERR "mach64_detect: Unknown ram size (%d) - using 8MB\n", card->mem_cntl & 0xf);
                card->Ram_Size = Mach64_Chip_Info.ramsize = 8 MB;
                break;
        } /* end switch */

    } else {

        switch( card->mem_cntl & 0x7 ) {
            case 0: card->Ram_Size = Mach64_Chip_Info.ramsize = 512 KB; break;
            case 1: card->Ram_Size = Mach64_Chip_Info.ramsize = 1 MB; break;
            case 2: card->Ram_Size = Mach64_Chip_Info.ramsize = 2 MB; break;
            case 3: card->Ram_Size = Mach64_Chip_Info.ramsize = 4 MB; break;
            case 4: card->Ram_Size = Mach64_Chip_Info.ramsize = 6 MB; break;
            case 5: card->Ram_Size = Mach64_Chip_Info.ramsize = 8 MB; break;
            default:
                printk( KERN_ERR "mach64_detect: Unknown ram size (%d) - using 8MB\n", card->mem_cntl & 0x7);
                card->Ram_Size = Mach64_Chip_Info.ramsize = 8 MB;
                break;
        } /* end switch */

    } /* end if */

} /* get_ramsize() */

/*
 *  find out aperture size, and calc MMIO_Base.
 */

void detect_aperture ( struct kgi_display *dpy ) {
    int		temp;
    char 	bus, devfn;
    GET_PRIVATE;

    /* enable an 8MB linear framebuffer */
    temp = mach64_in32(Config_Cntl, dpy);
    mach64_out32(Config_Cntl, 0x2 | (1<<19) | temp , dpy);

    pcibios_read_config_dword( bus, devfn, PCI_BASE_ADDRESS_0, &temp );

    
    Mach64_LFB_Mem.base = card->LFB_Base;
    if ( EOK == mem_check_region(&Mach64_LFB_Mem))
    {
   	 if (mem_claim_region(&Mach64_LFB_Mem))
	 printk("Claimed framebuffer memory:\n\
virtual address: 0x%8x\n\
physical base address: 0x%8x\n\
size: %d MB\n\
phys address lines: 0x%8x\n",
		(kgiu32) Mach64_LFB_Mem.baseptr, Mach64_LFB_Mem.base,
		Mach64_LFB_Mem.size / (1 MB), Mach64_LFB_Mem.decode);
	else
		DEBUG1("Error! could not claim framebuffer memory!");
	/* Set LFB_VBase to the _virtual_ address; this is what we'll
	  use from now on. */ 
	card->LFB_Size   = Mach64_LFB_Mem.size;
	card->LFB_VBase  = Mach64_LFB_Mem.baseptr;
	card->MMIO_Base  = (void*)(card->LFB_VBase + (8 MB) - (1 KB));
	card->Registers  = (int*)&Mach64_MMIO_Registers;
	card->Using_MMIO = 2;

	/* tell fbcon about our framebuffer */
	graphic_mmio_linear_fb.pstart = Mach64_LFB_Mem.base;
	graphic_mmio_linear_fb.psize =  
			graphic_mmio_linear_fb.size = card->Ram_Size;
	graphic_mmio_linear_fb.offset = Mach64_LFB_Mem.baseptr;
	
    	printk( KERN_ERR "mach64_detect: Aperture size: %d MB.\n", 
		    			card->LFB_Size / ( 1 MB ));
    	printk( KERN_ERR "mach64_detect: Aperture location (physical): 0x%8x.\n"
					, graphic_mmio_linear_fb.pstart);
		    			
	printk( KERN_ERR "mach64_detect: MMIO_Base: 0x%8x RAM: %d Kb.\n",
		(kgiu32)card->MMIO_Base, card->Ram_Size / 1024 );
		

    	/* From now on, we can access registers through the 
	   				big linear aperture. */
    }
    else
	DEBUG1("ERROR! could not claim framebuffer memory!");
	
}


/*
 * determine dac type
 */

void detect_dac( struct kgi_display *dpy ) {
    GET_PRIVATE;
    /* This is incorrect for 3d Rage cards. */
    card->Dac_Type	= ( card->dac_cntl & 0x70000 ) >> 16;

    printk( KERN_ERR "mach64_detect: Dac Type `%s'.\n",
			Mach64_RAMDAC_Descriptions[ card->Dac_Type ] );

}

#endif
